import 'dart:async';

import 'package:bloc/bloc.dart';
import 'package:equatable/equatable.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter_blue_plus_windows/flutter_blue_plus_windows.dart';
import 'package:objectbox/objectbox.dart';
import 'package:xiao_pet_tracker/bootstrap.dart';
import 'package:xiao_pet_tracker/objectbox.dart';
import 'package:xiao_pet_tracker/xiao_connector/models/capture_point.dart';

part 'xiao_connector_state.dart';

class XiaoConnectorCubit extends Cubit<XiaoConnectorState> {
  XiaoConnectorCubit() : super(const XiaoConnectorState());

  late BluetoothDevice device;
  List<BluetoothDevice> _systemDevices = [];
  List<BluetoothService> _services = [];

  late StreamSubscription<List<int>> dataCapturingSubscription;

  final ObjectBox _objectBox = getIt<ObjectBox>();
  late final Box<CapturePoint> _capturePointsBox;

  bool gotRotationX = false;
  bool gotRotationY = false;
  bool gotRotationZ = false;
  bool gotAcceleration = false;

  late String _rotX;
  late String _rotY;
  late String _rotZ;
  late String _accel;

  Future<void> init() async {
    _capturePointsBox = _objectBox.store.box<CapturePoint>();
  }

  Future<void> connect() async {
    emit(state.copyWith(status: XiaoConnectorStatus.loading));
    try {
      _systemDevices = await FlutterBluePlus.systemDevices([]);
    } catch (e) {
      emit(state.copyWith(status: XiaoConnectorStatus.failure));
      throw Exception('System Devices Error: $e');
    }

    try {
      // scan for devices for 15 seconds
      await FlutterBluePlus.startScan(timeout: const Duration(seconds: 15));
    } catch (e) {
      emit(state.copyWith(status: XiaoConnectorStatus.failure));
      throw Exception('Start Scan Error: $e');
    }

    FlutterBluePlus.scanResults.listen(
      (results) async {
        for (final r in results) {
          if (kDebugMode) {
            print(r);
          }

          // set the xiao sense name
          if (r.device.advName == 'Go Bluetooth') {
            await FlutterBluePlus.stopScan();
            device = r.device;

            await device.connect();

            try {
              _services = await device.discoverServices();
            } catch (e) {
              emit(state.copyWith(status: XiaoConnectorStatus.failure));
              throw Exception('Discover Services Error: $e');
            }

            device.connectionState
                .listen((BluetoothConnectionState blueState) async {
              if (blueState == BluetoothConnectionState.disconnected) {
                debugPrint('DISCONNECTED!');
                emit(state.copyWith(status: XiaoConnectorStatus.initial));
              }
            });

            // device.cancelWhenDisconnected(deviceConnectionStream);
            emit(state.copyWith(status: XiaoConnectorStatus.connected));
          }
        }
      },
      onDone: () {
        debugPrint('DONE');
        emit(state.copyWith(status: XiaoConnectorStatus.initial));
      },
    );
  }

  Future<void> startCapturing({required String captureType}) async {
    final senseService = _services
        .where(
          (s) => s.serviceUuid == Guid('4c534d36-4453-3354-5253-657276696365'),
        )
        .first
        .characteristics
        .where(
          (c) =>
              c.characteristicUuid ==
              Guid('61636365-6c65-7261-7469-6f6e44617461'),
        )
        .first;

    dataCapturingSubscription = senseService.onValueReceived.listen((value) {
      debugPrint('UPDATE: $value');
      debugPrint('UPDATE: ${String.fromCharCodes(value)}');

      final valueAsString = String.fromCharCodes(value);

      if (valueAsString.startsWith('rX=')) {
        gotRotationX = true;
        _rotX = valueAsString.replaceAll('rX=', '');
      }

      if (valueAsString.startsWith('rY=')) {
        gotRotationY = true;
        _rotY = valueAsString.replaceAll('rY=', '');
      }

      if (valueAsString.startsWith('rZ=')) {
        gotRotationZ = true;
        _rotZ = valueAsString.replaceAll('rZ=', '');
      }

      if (valueAsString.contains(',')) {
        gotAcceleration = true;
        _accel = valueAsString;
      }

      if (gotAcceleration && gotRotationX && gotRotationY && gotRotationZ) {
        final accelerations = _accel.split(',');
        final capturePoint = CapturePoint(
          type: captureType,
          rotationX: _rotX,
          rotationY: _rotY,
          rotationZ: _rotZ,
          accelerationX: accelerations[0],
          accelerationY: accelerations[1],
          accelerationZ: accelerations[2],
          millisecondsSinceEpoch: DateTime.now().toUtc().millisecondsSinceEpoch,
        );
        emit(state.copyWith(lastCapturedPoint: capturePoint));
        debugPrint("UPDATED POINT!");
        // _writeToObjectBox(capturePoint: capturePoint);
        gotRotationX = false;
        gotRotationY = false;
        gotRotationZ = false;
        gotAcceleration = false;
      }
    });

    device.cancelWhenDisconnected(dataCapturingSubscription);

    await senseService.setNotifyValue(true);

    emit(state.copyWith(status: XiaoConnectorStatus.capturing));
  }

  Future<void> stopCapturing() async {
    await dataCapturingSubscription.cancel();

    emit(state.copyWith(status: XiaoConnectorStatus.connected));
  }

  void _writeToObjectBox({
    required CapturePoint capturePoint,
  }) {
    _capturePointsBox.put(capturePoint);
  }

  void readFromObjectBox() {
    final points = _capturePointsBox.getAll();
    for (var i = 0; i < points.length; i++) {
      final point = points[i];
      debugPrint('Type: ${point.type}');
      debugPrint('Id: ${point.id}');
      debugPrint('RotX: ${point.rotationX}');
      debugPrint('RotY: ${point.rotationY}');
      debugPrint('RotZ: ${point.rotationZ}');
      debugPrint('AccelX: ${point.accelerationX}');
      debugPrint('AccelY: ${point.accelerationY}');
      debugPrint('AccelZ: ${point.accelerationZ}');
      debugPrint('Date: ${point.millisecondsSinceEpoch}');
    }
  }
}
